#include-once

#include <WinAPIEx.au3>
#include <APIConstants.au3>
#include <WindowsConstants.au3>
#include <GUIConstantsEx.au3>

; #INDEX# =======================================================================================================================
; Title .........: TrayIconEx_UDF
; UDF Version ...: 1.0.0.0
; AutoIt Version : 3.3.8.0
; Description ...: Creates, deletes and manages tray icons.
; Author(s) .....: FireFox (aka d3mon, d3monCorp)
; Dll ...........: None
; ===============================================================================================================================

; #VARIABLES# ===================================================================================================================
Global $_aTrayIconEx_Data[11][4] = [[0]]
Global Enum $_iTrayIconEx_Icon, $_iTrayIconEx_CallBack, $_iTrayIconEx_IsFlashing, $_iTrayIconEx_FlashingState
Global $_iTrayLastId = 0
; ===============================================================================================================================

; #CONSTANTS# ===================================================================================================================
Local Const $TM_HANDLER = $WM_USER + 1

Global Const $TRAYICONSTATE_SHOW = 1, $TRAYICONSTATE_HIDE = 2, $TRAYICONSTATE_FLASH = 4, $TRAYICONSTATE_STOPFLASH = 8
Global Const $TRAYICON_FLASHTIMEOUT = 750

Global Const $_sTrayIconEx_DefaultToolTip = "AutoIt - " & @ScriptName
; ===============================================================================================================================

; #INIT# ========================================================================================================================
Global Const $_hTrayIconEx_WndMsg = GUICreate("TrayIconEx")
GUIRegisterMsg($TM_HANDLER, "__TrayIconEx_TMHANDLER")
GUIRegisterMsg($WM_TIMER, "__TrayIconEx_WMTIMER")

OnAutoItExitRegister("__TrayIconEx_OnAutoItExit")
; ===============================================================================================================================

; #CURRENT# =====================================================================================================================
;_TrayIconEx_Create
;_TrayIconEx_Delete
;_TrayIconEx_SetIcon
;_TrayIconEx_SetState
;_TrayIconEx_SetToolTip
;_TrayIconEx_SetTip
; ===============================================================================================================================

; #INTERNAL_USE_ONLY# ===========================================================================================================
;__TrayIconEx_TMHANDLER
;__TrayIconEx_WMTIMER
;__TrayIconEx_GetIconHandle
;__TrayIconEx_GetFreeSlot
;__TrayIconEx_OnAutoItExit
; ===============================================================================================================================

; #FUNCTION# ====================================================================================================================
; Name...........: _TrayIconEx_Create
; Description ...: Creates a tray icon (not visible).
; Syntax.........: _TrayIconEx_Create($sIconFile = Default, $iIconID = Default, $sCallback = "", $hIcon = 0)
; Parameters ....: $sIconFile	- if Default, @AutoItExe is set; File path that contains an icon.
;				   $iIconID		- if Default, 0 is set; Icon index of the file.
;				   $sCallback	- Callback function to be called on a mouse event. This last must have the following two
;								  parameters: TrayId for the tray icon id returned by the function _TrayIconEx_Create and
;								  MouseEvent for the mouse event message.
;				   $hIcon		- Icon handle, if set the first two parameters will be ignored.
; Return values .: Success		- Tray icon Id.
;				   Failure		- huh, autoit crash.
; Author ........: FireFox
; Modified.......:
; Remarks .......: This function does not create a tray icon structure, it only saves some data to be used when the function
; _TrayIconEx_SetState is called, so you must call this last before any other function (except _TrayIconEx_Delete).
; Related .......: _TrayIconEx_Delete
; Link ..........:
; Example .......: Yes
; ===============================================================================================================================
Func _TrayIconEx_Create($sIconFile = Default, $iIconID = Default, $sCallback = "", $hIcon = 0)
	If $sIconFile = Default Then $sIconFile = @AutoItExe
	If $iIconID = Default Then $iIconID = 0
	If $hIcon = 0 Then $hIcon = __TrayIconEx_GetIconHandle($sIconFile, $iIconID)

	Local $iTrayId = __TrayIconEx_GetFreeSlot()

	$_aTrayIconEx_Data[$iTrayId][$_iTrayIconEx_Icon] = $hIcon
	$_aTrayIconEx_Data[$iTrayId][$_iTrayIconEx_CallBack] = $sCallback
	$_aTrayIconEx_Data[$iTrayId][$_iTrayIconEx_IsFlashing] = False
	$_aTrayIconEx_Data[$iTrayId][$_iTrayIconEx_FlashingState] = False

	$_aTrayIconEx_Data[0][0] += 1

	$_iTrayLastId = $iTrayId

	Return $iTrayId
EndFunc   ;==>_TrayIconEx_Create

; #FUNCTION# ====================================================================================================================
; Name...........: _TrayIconEx_Delete
; Description ...: Deletes a tray icon.
; Syntax.........: _TrayIconEx_Delete($iTrayId = -1)
; Parameters ....: $iTrayId		- if -1 the last tray icon will be taken; tray icon Id to delete.
; Return values .: Success      - True
;				   Failure      - huh, autoit crash.
; Author ........: FireFox
; Modified.......:
; Remarks .......:
; Related .......: _TrayIconEx_Create
; Link ..........:
; Example .......: Yes
; ===============================================================================================================================
Func _TrayIconEx_Delete($iTrayId = -1)
	If $iTrayId = -1 Then $iTrayId = $_iTrayLastId

	Local $tNID = DllStructCreate($tagNOTIFYICONDATA)

	DllStructSetData($tNID, "Size", DllStructGetSize($tNID))
	DllStructSetData($tNID, "hWnd", $_hTrayIconEx_WndMsg)
	DllStructSetData($tNID, "ID", $iTrayId)

	Local $blRet = _WinAPI_ShellNotifyIcon($NIM_DELETE, $tNID)
	If Not $blRet Then Return False

	_WinAPI_DestroyIcon($_aTrayIconEx_Data[$iTrayId][$_iTrayIconEx_Icon])

	For $iTrayDataSub = 0 To 3
		$_aTrayIconEx_Data[$iTrayId][$iTrayDataSub] = 0
	Next

	$_aTrayIconEx_Data[0][0] -= 1

	Return True
EndFunc   ;==>_TrayIconEx_Delete

; #FUNCTION# ====================================================================================================================
; Name...........: _TrayIconEx_SetIcon
; Description ...: Sets the icon to a tray.
; Syntax.........: _TrayIconEx_SetIcon($iTrayId = -1, $sIconFile = @AutoItExe, $iIconID = 0, $hIcon = 0, $blDestroyIcon = True,
;				   $blReplaceIconHandle = True)
; Parameters ....: $iTrayId		- if -1 the last tray icon will be taken; tray icon Id to delete.
;				   $sIconFile	- if Default, @AutoItExe is set; File path that contains an icon.
;				   $iIconID		- if Default, 0 is set; Icon index of the file.
;				   $hIcon		- Icon handle, if set the parameters 2 and 3 will be ignored.
;				   $blDestroyIcon	- Destroy or not the previous icon ressource (internal use).
;				   $blReplaceIconHandle	- Replaces or not the previous icon handle (internal use).
; Return values .: Success		- True
;				   Failure		- False
; Author ........: FireFox
; Modified.......:
; Remarks .......:
; Related .......: _TrayIconEx_Create
; Link ..........:
; Example .......: Yes
; ===============================================================================================================================
Func _TrayIconEx_SetIcon($iTrayId = -1, $sIconFile = @AutoItExe, $iIconID = 0, $hIcon = 0, $blDestroyIcon = True, $blReplaceIconHandle = True)
	If $iTrayId = -1 Then $iTrayId = $_iTrayLastId

	If $hIcon = 0 Then $hIcon = __TrayIconEx_GetIconHandle($sIconFile, $iIconID)

	Local $tNID = DllStructCreate($tagNOTIFYICONDATA)

	DllStructSetData($tNID, "Size", DllStructGetSize($tNID))
	DllStructSetData($tNID, "hWnd", $_hTrayIconEx_WndMsg)
	DllStructSetData($tNID, "ID", $iTrayId)
	DllStructSetData($tNID, "Flags", $NIF_ICON)
	DllStructSetData($tNID, "hIcon", $hIcon)

	If $blDestroyIcon Then _WinAPI_DestroyIcon($_aTrayIconEx_Data[$iTrayId][$_iTrayIconEx_Icon])

	Local $blRet = _WinAPI_ShellNotifyIcon($NIM_MODIFY, $tNID)
	If $blRet Then
		If $blReplaceIconHandle Then $_aTrayIconEx_Data[$iTrayId][$_iTrayIconEx_Icon] = $hIcon
	Else
		_WinAPI_DestroyIcon($hIcon)
		$_aTrayIconEx_Data[$iTrayId][$_iTrayIconEx_Icon] = 0
	EndIf

	Return $blRet
EndFunc   ;==>_TrayIconEx_SetIcon

; #FUNCTION# ====================================================================================================================
; Name...........: _TrayIconEx_SetState
; Description ...: Sets the state of a tray icon.
; Syntax.........: _TrayIconEx_SetState($iTrayId = -1, $iTrayState = $TRAYICONSTATE_SHOW)
; Parameters ....: $iTrayId		- if -1 the last tray icon will be taken; tray icon Id to delete.
;				   $iTrayState	- tray state, it can take the following values:
;								  |$TRAYICONSTATE_SHOW	- 1
;								  |$TRAYICONSTATE_HIDE	- 2
;								  |$TRAYICONSTATE_FLASH - 4
;								  |$TRAYICONSTATE_STOPFLASH	- 8
; Return values .: Success		- True
;				   Failure		- False
; Author ........: FireFox
; Modified.......:
; Remarks .......:
; Related .......: _TrayIconEx_Create
; Link ..........:
; Example .......: Yes
; ===============================================================================================================================
Func _TrayIconEx_SetState($iTrayId = -1, $iTrayState = $TRAYICONSTATE_SHOW)
	If $iTrayId = -1 Then $iTrayId = $_iTrayLastId

	Local $tNID = DllStructCreate($tagNOTIFYICONDATA), $blRet = False

	If BitAND($iTrayState, $TRAYICONSTATE_SHOW) Then
		DllStructSetData($tNID, "Size", DllStructGetSize($tNID))
		DllStructSetData($tNID, "hWnd", $_hTrayIconEx_WndMsg)
		DllStructSetData($tNID, "ID", $iTrayId)
		DllStructSetData($tNID, "Flags", BitOR($NIF_ICON, $NIF_MESSAGE))
		DllStructSetData($tNID, "CallbackMessage", $TM_HANDLER)
		DllStructSetData($tNID, "hIcon", $_aTrayIconEx_Data[$iTrayId][$_iTrayIconEx_Icon])

		$blRet = _WinAPI_ShellNotifyIcon($NIM_ADD, $tNID)
		If Not $blRet Then Return False
	ElseIf BitAND($iTrayState, $TRAYICONSTATE_HIDE) Then
		DllStructSetData($tNID, "Size", DllStructGetSize($tNID))
		DllStructSetData($tNID, "hWnd", $_hTrayIconEx_WndMsg)
		DllStructSetData($tNID, "ID", $iTrayId)

		$blRet = _WinAPI_ShellNotifyIcon($NIM_DELETE, $tNID)
		If Not $blRet Then Return False
	EndIf

	If BitAND($iTrayState, $TRAYICONSTATE_FLASH) Then
		If Not $_aTrayIconEx_Data[$iTrayId][$_iTrayIconEx_IsFlashing] Then
			_WinAPI_SetTimer($_hTrayIconEx_WndMsg, $iTrayId, $TRAYICON_FLASHTIMEOUT, 0)
			If @error Then Return False

			$_aTrayIconEx_Data[$iTrayId][$_iTrayIconEx_IsFlashing] = True

			$blRet = True
		EndIf
	ElseIf BitAND($iTrayState, $TRAYICONSTATE_STOPFLASH) Then
		If $_aTrayIconEx_Data[$iTrayId][$_iTrayIconEx_IsFlashing] Then
			_WinAPI_KillTimer($_hTrayIconEx_WndMsg, $iTrayId)

			$_aTrayIconEx_Data[$iTrayId][$_iTrayIconEx_IsFlashing] = False
		EndIf

		DllStructSetData($tNID, "Size", DllStructGetSize($tNID))
		DllStructSetData($tNID, "hWnd", $_hTrayIconEx_WndMsg)
		DllStructSetData($tNID, "ID", $iTrayId)
		DllStructSetData($tNID, "Flags", $NIF_ICON)
		DllStructSetData($tNID, "hIcon", $_aTrayIconEx_Data[$iTrayId][$_iTrayIconEx_Icon])

		$blRet = _WinAPI_ShellNotifyIcon($NIM_MODIFY, DllStructGetPtr($tNID))
	EndIf

	Return $blRet
EndFunc   ;==>_TrayIconEx_SetState

; #FUNCTION# ====================================================================================================================
; Name...........: _TrayIconEx_SetToolTip
; Description ...: Sets a tool tip to a tray icon.
; Syntax.........: _TrayIconEx_SetToolTip($iTrayId = -1, $sToolTip = $_sTrayIconEx_DefaultToolTip)
; Parameters ....: $iTrayId		- if -1 the last tray icon will be taken; tray icon Id to delete.
;				   $sToolTip	- tool tip text to set, if omitted the $_sTrayIconEx_DefaultToolTip value will be taken.
; Return values .: Success		- True
;				   Failure		- False
; Author ........: FireFox
; Modified.......:
; Remarks .......:
; Related .......: _TrayIconEx_Create
; Link ..........:
; Example .......: Yes
; ===============================================================================================================================
Func _TrayIconEx_SetToolTip($iTrayId = -1, $sToolTip = $_sTrayIconEx_DefaultToolTip)
	If $iTrayId = -1 Then $iTrayId = $_iTrayLastId

	Local $tNID = DllStructCreate($tagNOTIFYICONDATA)

	DllStructSetData($tNID, "Size", DllStructGetSize($tNID))
	DllStructSetData($tNID, "hWnd", $_hTrayIconEx_WndMsg)
	DllStructSetData($tNID, "ID", $iTrayId)
	DllStructSetData($tNID, "Flags", $NIF_TIP)
	DllStructSetData($tNID, "Tip", $sToolTip)

	Return _WinAPI_ShellNotifyIcon($NIM_MODIFY, $tNID)
EndFunc   ;==>_TrayIconEx_SetToolTip

; #FUNCTION# ====================================================================================================================
; Name...........: _TrayIconEx_SetTip
; Description ...: Sets a tip to a tray icon.
; Syntax.........: _TrayIconEx_SetTip($iTrayId = -1, $sTitle = "Title", $sText = "Text", $iTimeOut = 10, $iInfoFlags = 0)
; Parameters ....: $iTrayId		- if -1 the last tray icon will be taken; tray icon Id to delete.
;				   $sTitle		- Title to set, if omitted "Title" will be taken.
;				   $sText		- Text to set, if omitted, "Text" will be taken.
;				   $iTimeOut	- Tip display timeout, if omitted 10 will be taken (10 secs).
;				   $iInfoFlags	- Icon flag, check the option param of the TrayTip function to know its possible values.
; Return values .: Success		- True
;				   Failure		- False
; Author ........: FireFox
; Modified.......:
; Remarks .......: Not compatible with non WIN32_NT OS type and WIN_NT4 OS version.
; Related .......: _TrayIconEx_Create
; Link ..........:
; Example .......: Yes
; ===============================================================================================================================
Func _TrayIconEx_SetTip($iTrayId = -1, $sTitle = "Title", $sText = "Text", $iTimeOut = 10, $iInfoFlags = 0)
	If @OSType <> "WIN32_NT" Or @OSVersion = "WIN_NT4" Then Return False

	If $iTrayId = -1 Then $iTrayId = $_iTrayLastId

	Local $tNID = DllStructCreate($tagNOTIFYICONDATA)

	DllStructSetData($tNID, "Size", DllStructGetSize($tNID))
	DllStructSetData($tNID, "hWnd", $_hTrayIconEx_WndMsg)
	DllStructSetData($tNID, "ID", $iTrayId)
	DllStructSetData($tNID, "Flags", $NIF_INFO)
	DllStructSetData($tNID, "Info", $sText)
	DllStructSetData($tNID, "Version", $iTimeOut * 1000)
	DllStructSetData($tNID, "InfoTitle", $sTitle)
	DllStructSetData($tNID, "InfoFlags", $iInfoFlags)

	Return _WinAPI_ShellNotifyIcon($NIM_MODIFY, $tNID)
EndFunc   ;==>_TrayTip

Func __TrayIconEx_TMHANDLER($hWnd, $Msg, $wParam, $lParam)
	Switch $hWnd
		Case $_hTrayIconEx_WndMsg
			Local $iTrayId = Number($wParam)

			If $_aTrayIconEx_Data[$iTrayId][1] <> "" Then
				Call($_aTrayIconEx_Data[$iTrayId][1], $iTrayId, Number($lParam))
			EndIf
	EndSwitch

	Return $GUI_RUNDEFMSG
EndFunc   ;==>__TrayIconEx_WndProc

Func __TrayIconEx_WMTIMER($hWnd, $Msg, $wParam, $lParam)
	Local $iTrayId = Number($wParam)

	If $_aTrayIconEx_Data[$iTrayId][$_iTrayIconEx_FlashingState] Then
		_TrayIconEx_SetIcon($iTrayId, "", 0, $_aTrayIconEx_Data[$iTrayId][$_iTrayIconEx_Icon], False, False)
	Else
		_TrayIconEx_SetIcon($iTrayId, "", 0, 0, False, False)
	EndIf

	$_aTrayIconEx_Data[$iTrayId][$_iTrayIconEx_FlashingState] = Not $_aTrayIconEx_Data[$iTrayId][$_iTrayIconEx_FlashingState]

	Return $GUI_RUNDEFMSG
EndFunc   ;==>__TrayIconEx_WMTIMER

Func __TrayIconEx_GetIconHandle($sIconFile, $iIconID)
	Local $hIcon = 0

	$iIconID = _GetIconID($iIconID, $sIconFile)

	If $sIconFile = "" Then
		$hIcon = _WinAPI_LoadIcon(0, $iIconID)
	Else
		$hIcon = _WinAPI_ExtractIcon($sIconFile, $iIconID, True)
		If $hIcon = 0 Then
			$hIcon = _WinAPI_LoadIcon(0, $IDI_ASTERISK)
		EndIf
	EndIf

	Return $hIcon
EndFunc   ;==>__TrayIconEx_GetIconHandle

Func __TrayIconEx_GetFreeSlot()
	For $iTrayId = 1 To UBound($_aTrayIconEx_Data) -1
		If $_aTrayIconEx_Data[$iTrayId][$_iTrayIconEx_Icon] = 0 Then
			Return $iTrayId
		EndIf
	Next

	If $_aTrayIconEx_Data[0][0] > 0 And Mod($_aTrayIconEx_Data[0][0], 10) = 0 Then
		ReDim $_aTrayIconEx_Data[$_aTrayIconEx_Data[0][0] + 11][4]
	EndIf

	Return $_aTrayIconEx_Data[0][0] + 1
EndFunc   ;==>__TrayIconEx_GetFreeSlot

Func __TrayIconEx_OnAutoItExit()
	For $iTrayId = 1 To UBound($_aTrayIconEx_Data) -1
		If $_aTrayIconEx_Data[$iTrayId][$_iTrayIconEx_Icon] Then
			_TrayIconEx_Delete($iTrayId)
		EndIf
	Next
	$_aTrayIconEx_Data = 0

	GUIDelete($_hTrayIconEx_WndMsg)
EndFunc   ;==>__TrayIconEx_OnAutoItExit

Func _GetIconID($iID, $sFile)
	If StringRight($sFile, 4) = ".exe" Then
		If $iID < 0 Then
			$iID = -($iID + 1)
		ElseIf $iID > 0 Then
			$iID = -$iID
		EndIf
	ElseIf StringRight($sFile, 4) = ".icl" And $iID < 0 Then
		$iID = -($iID + 1)
	Else
		If $iID > 0 Then
			$iID = -$iID
		ElseIf $iID < 0 Then
			$iID = -($iID + 1)
		EndIf
	EndIf

	Return $iID
EndFunc   ;==>_GetIconID